class AardvarkException(Exception):
    pass


class AccessAdvisorException(AardvarkException):
    pass


class CombineException(AardvarkException):
    pass


class DatabaseException(AardvarkException):
    pass


class RetrieverException(AardvarkException):
    pass
